(* types of Godel's System T *)
type tp = Bool 
        | Nat 
        | Arr of (tp * tp)   (* Arr (ty1,ty2) = ty1 -> ty2 *)
        | Times of (tp * tp) (* Times (ty1,ty2) = ty1 * ty2 *)
;;

(* terms of Godel's System T *)
(* non usiamo nomi di variabile ma *indici*
 * (codifica di de Bruijn)
 * Rel i è un indice che punta alla (i+1)-esima lambda
 * più esterna
 * Ad esempio 
 * \f:N->N->N.\x,y:N.f x y
 * =
 * Lam (Arr (Nat, Arr (Nat,Nat)), (*f*)
 *   Lam (Nat, (*x*)
 *     Lam (Nat, (*y*)
 *       App (App (Rel 2, (*f*)
 *                 Rel 1), (*x*)
 *            Rel 0)))) (*y*)
 *)
type tm = Zero                   (* zero *)
        | Succ of tm             (* successor *)
        | Bval of bool           (* boolean values *)
        | Rel of int             (* de Bruijn index *)
        | Lam of (tp * tm)       (* lambda abstraction *)
        | App of (tm * tm)       (* application *)
        | Pair of (tm * tm)      (* pair *)
        | Fst of tm              (* first of a pair *)
        | Snd of tm              (* second of a pair *)
        | Disc of (tm * tm * tm) (* if-then-else *)
        | Rec of (tm * tm * tm)  (* recursion operator *)
;;

(* typechecker
 * the context ctx is a list containing type declarations for
 * "dangling" indexes, i.e. indexes whose value is "too big"
 * to point to a lambda
 * dangling indexes correspond to free variables
 * (variables that are not bound by any lambda)
 *)
let rec typecheck_T ctx = function
  | Zero -> Nat
  | Succ n -> (match typecheck_T ctx n with
    | Nat -> Nat
    | _ -> failwith "typechecking error")
  | Bval _ -> Bool
  | Rel i -> (try (List.nth ctx i) with
              | _ -> failwith "typechecking error")
  | Lam (src,body) -> Arr (src, typecheck_T (src::ctx) body)
  | App (t,u) -> 
     (let u_ty = typecheck_T ctx u in
       match typecheck_T ctx t with
       | Arr (src,tgt) when (src = u_ty) -> tgt
       | _ -> failwith "typechecking error")
  | Pair (a,b) -> 
     let a_ty = typecheck_T ctx a in
     let b_ty = typecheck_T ctx b in
     Times(a_ty,b_ty)
  | Fst p -> (match typecheck_T ctx p with
    | Times (a,b) -> a
    | _ -> failwith "typechecking error")
  | Snd p -> (match typecheck_T ctx p with
    | Times (a,b) -> b
    | _ -> failwith "typechecking error")
  | Disc (u,v,b) -> 
     let u_ty = typecheck_T ctx u in
     let v_ty = typecheck_T ctx v in
     let b_ty = typecheck_T ctx b in
     if (u_ty = v_ty && b_ty = Bool)
       then u_ty
       else failwith "typechecking error"
  | Rec (u,v,n) ->  
     let u_ty = typecheck_T ctx u in
     let v_ty = typecheck_T ctx v in
     let n_ty = typecheck_T ctx n in
     if n_ty = Nat then
        match v_ty with
        | Arr (ty1, Arr (Nat, ty2)) 
            when ty1 = u_ty && ty2 = u_ty ->
              u_ty
        | _ -> failwith "typechecking error"
     else failwith "typechecking error"
;;

(* sum *)
let plus_T =
  (* \x,y:nat.Rec x (\z,x'.Succ z) y *)
  Lam (Nat, (*x*)
    Lam (Nat, (*y*)
      Rec (Rel 1, (*x*)
        Lam (Nat, (*z*)
          Lam (Nat, (*x'*)
            Succ (Rel 1) (*z*)
        )), Rel 0))) (*y*)
;;

(* product *)
let times_T =
  (* \x,y:nat.Rec Zero (\z,x'.plus_T z x) y *)
  Lam (Nat, (*x*)
    Lam (Nat, (*y*)
      Rec (Zero, (*x*)
        Lam (Nat, (*z*)
          Lam (Nat, (*x'*)
            App ( App (plus_T,
              Rel 1), (*z*)
            Rel 3))), (*x*)
        Rel 0))) (*y*)
;;

(* \f:N->N.\n:nat.Ite 1 f n
 * i.e.
 * \f:N->N.\n:nat.Rec 1 (\a,b:nat.f a) n
 * of type (nat -> nat) -> (nat -> nat)
 *)
let next_fun = 
  Lam (Arr (Nat,Nat), (*f*)
    Lam (Nat, (*n*)
      Rec (Succ Zero,
           Lam (Nat, (*a*)
             Lam (Nat, (*b*)
               App (Rel 3, (*f*)
                    Rel 1))), (*a*)
           Rel 0))) (*n*)
;;

(* \z,y:nat.Ite (times_T y) next z
 * i.e.
 * \z,y:nat.Rec (times_T y) (\f:nat->nat.\r:nat.next_fun f) z
 * of type nat -> nat -> nat -> nat
 *)
let ack2_T =
  Lam (Nat, (*z*)
    Lam (Nat, (*y*)
      Rec (App (times_T,Rel 0), (*y*)
           Lam (Arr (Nat,Nat), (*f*)
             Lam (Nat, (*r*)
               App (next_fun, Rel 1))), (*f*)
           Rel 1))) (*z*)
;;

(* ackermann function *)
let ack_T =
  Lam (Nat, (*z*)
    Rec (plus_T,
         Lam (Arr (Nat, Arr (Nat,Nat)), (*f*)
           Lam (Nat, (*w*)
             App (ack2_T, Rel 0))), (*w*)
         Rel 0)) (*z*)
;;

(* lift increments dangling indexes when needed *)
let rec lift t n k =
  match t with
  | Zero | Bval _ -> t
  | Succ u -> Succ (lift u n k)
  | Rel i when k <= i -> Rel (i+n)
  | Rel _ -> t
  | Pair (u,v) -> Pair (lift u n k, lift v n k)
  | Fst u -> Fst (lift u n k)
  | Snd u -> Snd (lift u n k)
  | Disc (u,v,b) -> Disc (lift u n k, lift v n k, lift b n k)
  | Rec (u,v,m) -> Rec (lift u n k, lift v n k, lift m n k)
  | Lam (s,u) -> Lam (s,lift u n (k+1))
  | App (u,v) -> App (lift u n k, lift v n k)
;;

(* substitution t [ u / Rel 0 ] *)
let subst t u =
  let rec aux t' k =
    match t' with
    | Zero | Bval _ -> t'
    | Succ t0 -> Succ (aux t0 k)
    | Rel i when i = k -> lift u k 0
    | Rel i when i < k -> t'
    | Rel i -> Rel (i-1)
    | Pair (t0,t1) -> Pair (aux t0 k,aux t1 k)
    | App (t0,t1) -> App (aux t0 k,aux t1 k)
    | Lam (s,t0) -> Lam (s,aux t0 (k+1))
    | Fst t0 -> Fst (aux t0 k)
    | Snd t0 -> Snd (aux t0 k)
    | Disc (t0, t1, t2) -> Disc (aux t0 k, aux t1 k, aux t2 k)
    | Rec (t0,t1,t2) -> Rec (aux t0 k, aux t1 k, aux t2 k)
  in aux t 0
;;

(* roughly corresponding to 1-step parallel reduction *)
let rec reduce t =
  match t with
  | Zero | Bval _ | Rel _ -> t
  | Succ u -> Succ (reduce u)
  | App (Lam (_,u), v) -> subst u v
  | App (u,v) -> App (reduce u,reduce v)
  | Lam (s,u) -> Lam (s,reduce u)
  | Rec (b,f,Zero) -> b
  | Rec (b,f,Succ m) -> App (App (f,Rec (b,f,m)),m)
  | Rec (b,f,m) -> Rec (reduce b,reduce f,reduce m)
  | Pair (u,v) -> Pair (reduce u,reduce v)
  | Fst u -> Fst (reduce u)
  | Snd u -> Snd (reduce u)
  | Disc (u,v,Bval true) -> u
  | Disc (u,v,Bval false) -> v
  | Disc (u,v,b) -> Disc (reduce u, reduce v, reduce b)
;;

(* iterates reduce until the term doesn't reduce any more *)
let rec eval t =
  let t' = reduce t in
  if t = t' then t
  else eval t'
;;

let rec nat_of_t =
   function
   | Zero -> 0
   | Succ x -> 1 + (nat_of_t x)
   | _ -> failwith "not a nat"
  ;;

let rec t_of_nat n =
    assert (n >= 0);
    if n = 0 then Zero
    else Succ (t_of_nat (n-1))
  ;;

